#include "filters.h"
#include <assert.h>
#include <math.h>

Bin_Image *binarize(Image *img, int bin_type, int threshold)
{
    Gray_Image *gimg = apply_grayscale(img);
    Bin_Image *bimg;
    if (bin_type == OTSU1)
        bimg = apply_otsu(gimg);
    else if (bin_type == OTSU2)
        bimg = apply_otsu2(gimg);
    else
        bimg = apply_threshold(gimg, threshold);
    free_grayimage(gimg);
    return bimg;
}

Bin_Image *apply_threshold(Gray_Image *img, int threshold)
{
    Bin_Image *s = create_bin_image(img->w, img->h);

    for(USHORT x = 0; x < img->w; x++)
    {
        for(USHORT y = 0; y < img->h; y++)
        {
            BYTE px = getpixelgray(img, x, y);
            setpixelbin(s, x, y, px > threshold ? 0x00 : 0xFF);
        }
    }
    return s;
}

Bin_Image *apply_otsu(Gray_Image *img)
{
    int histogram[256] = {0};
    get_histogram(img, histogram);

    float sum = 0;
    for (int i = 0; i < 255; i++)
        sum += i * histogram[i];

    float sumB = 0, wB = 0, wF = 0, varmax = 0, threshold = 0;
    int pxnbr = img->w * img->h;

    for (int i = 0; i < 255; i++)
    {
        wB += (int)histogram[i];
        if (wB == 0)
            continue;

        wF = pxnbr - wB;
        if (wF == 0)
            break;

        sumB += (float)(i * histogram[i]);

        float mB = sumB / wB;
        float mF = (sum - sumB) / wF;

        float mBF = mB - mF;
        float variance = wB * wF * mBF * mBF;

        if (variance > varmax)
        {
            varmax = variance;
            threshold = i;
        }
    }

    return apply_threshold(img, (int)threshold);
}

Bin_Image *apply_otsu2(Gray_Image *img)
{
    int histogram[256] = {0};
    get_histogram(img, histogram);

    double w = 0.0;
    double u = 0.0;
    double uT = 0.0;

    int k = 256;
    int threshold = 0;

    double histNormalized[256];

    long double work1 = 0.0, work2 = 0.0;
    double work3 = 0.0;

    for (int i = 1; i <= k; ++i)
        histNormalized[i - 1] = histogram[i - 1] / (double)(img->w * img->h);

    for (int i = 1; i <= k; ++i)
        uT += i * histNormalized[i - 1];

    for (int i = 1; i <= k; ++i)
    {
        w += histNormalized[i - 1];
        u += i * histNormalized[i - 1];
        work1 = uT * w - u;
        work2 = (work1 * work1) / (w * (1.0f - w));
        if (work2 > work3)
            work3 = work2;
    }
    threshold = toint(sqrt(work3));
    return apply_threshold(img, threshold);


}

Gray_Image *apply_grayscale(Image* img)
{
    Gray_Image *s = create_gray_image(img->w, img->h);

    for(USHORT x = 0; x < img->w; x++)
    {
        for(USHORT y = 0; y < img->h; y++)
        {
            Pixel px = getpixel(img, x, y);
            BYTE gray = (BYTE)toint((px.red + px.green + px.blue) / 3);
            setpixelgray(s, x, y, gray);
        }
    }

    return s;
}

Image *apply_gaussblur(Image* img, USHORT length)
{

    Image *s = create_image(img->w, img->h);
    USHORT x;
    USHORT y;
    short mx;
    short my;
    float r, g, b, factor;
    Pixel current;
    Pixel sub;
    for(x = 0; x < img->w; x++)
    {
        for(y = 0; y < img->h; y++)
        {
            r = 0;
            g = 0;
            b = 0;
            current = getpixel(img, x, y);
            for(mx = x - length; mx < x + length; mx++)
            {
                for(my = y - length; my < y + length; my++)
                {
                    if(mx == x && my == y)
                    {
                        r += current.red;
                        g += current.green;
                        b += current.blue;
                    }
                    else
                    {
                        factor = 1 / ((float)(MAX(ABS(mx - x), ABS(my - y)) + 1) * 4 - 4);
                        sub = getpixel_s(img, mx, my, current);
                        r += sub.red * factor;
                        g += sub.green * factor;
                        b += sub.blue * factor;
                    }
                }
            }
            setpixelrgb(s, x, y, r / (2 * length), g / (2 * length), b / (2 * length));
        }
    }
    return s;
}

