#include "loadimg.h"
#include "imgmanip.h"
#include <unistd.h>
#include <errno.h>
#include <sys/wait.h>

///Internal fonctions
static Image *processbmpheader(_in_ BYTE* header, BYTE* bpp);
static bool loadbmp_1(_in_ FILE*, _inout_ Image*);
static bool loadbmp_8(_in_ FILE*, _inout_ Image*);
static bool loadbmp_16(_in_ FILE*, _inout_ Image*);
static bool loadbmp_24(_in_ FILE*, _inout_ Image*);
static bool loadbmp_32(_in_ FILE*, _inout_ Image*, bool);
static Image *load_image_internal(_in_ const char *path);
static bool canwrite(_in_ const char *);

bool boverwrite = false;

typedef struct //Bitmap header
{
    char bfType[2];
    int bfSize;
    int bfReserved;
    int bfOffBits;
    int biSize;
    int biWidth;
    int biHeight;
    short biPlanes;
    short biBitCount;
    int biCompression;
    int biSizeImage;
    int biXPelsPerMeter;
    int biYPelsPerMeter;
    int biClrUsed;
    int biClrImportant;
} BmpHeader;

static bool canwrite(const char *path)
{
    if(boverwrite)
        return true;

    FILE *f = fopen(path, "r");
    if(!f)
        return true;
    fclose(f);
    char c;
    fseek(stdin,0,SEEK_END);
    while(true)
    {
        printf("File : \"%s\" already exists, overwrite ? (y/n)\n", path);
        c = getc(stdin);
        if(c == 'n')
            return false;
        if(c == 'y')
            return true;
    }
}

Image *processbmpheader(BYTE *header, BYTE *bpp)
{
    if(*(int*)header == 40 || *(int*)header == 52 || *(int*)header == 56
            || *(int*)header == 108 || *(int*)header == 124)
    {
        if(*(int*)(header + 16) == 3)
            *bpp = 32;
        else if (*(int*)(header + 16) != 0)
        {
            printf("Warning : compressed bitmap are not supported");
            return NULL;
        }
        else
            *bpp = *(int*)(header + 14);
        Image *img = create_image(*(int*)(header + 4), *(int*)(header + 8));
        return img;
    }
    else if(*(int*)header == 14)
    {
        USHORT w, h;
        memcpy(&w, header + 2, 2);
        memcpy(&h, header + 2, 2);
        Image *img = create_image(*(int*)(header + 4), *(int*)(header + 8));
        *bpp = 24;
        return img;
    }
    return NULL;
}

Image *load_image(const char *path)
{
    Image *img = load_image_internal(path);
    if(img == NULL)
    {
        FILE *f=fopen(path,"r");
        if(f==NULL)
            return NULL;
        fclose(f);
        verbose("Failed to load :\"%s\", trying to convert file\n", path);
        pid_t pid = fork();
        int status;
        if(pid == 0)
        {
            if(execl("/usr/bin/convert", "convert", path , "-type", " truecolor", "tmp.bmp",
                     (char*)0) == -1)
            {
                exit(EXIT_FAILURE);
            }
            exit(EXIT_SUCCESS);
        }
        waitpid(pid, &status, 0);
        if(status != EXIT_SUCCESS)
        {
            verbose("Image conversion failed, %i\n", errno);
            return NULL;
        }
        img = load_image_internal("tmp.bmp");
        if(unlink("tmp.bmp") == -1)
            printf("Failed to remove temporary file : \"tmp.bmp\"");
    }
    return img;
}

static Image *load_image_internal(const char *path)
{
    FILE *file = fopen(path, "rb");
    if(!file)
    {
        verbose("Failed to open file : \"%s\"\n", path);
        return NULL;
    }
    verbose("Opening : \"%s\"\n", path);
    int offset = 0;
    BYTE buffer[40]; //40=max header size
    BYTE bpp;
    if(!fread(buffer, 14, 1, file))
    {
        fclose(file);
        return NULL;
    }
    if(buffer[0] != 'B' || buffer[1] != 'M')
    {
        verbose("Invalid header\n");
        fclose(file);
        return NULL;
    }
    memcpy(&offset, &buffer[10], sizeof(int));
    if(!fread(buffer, 40, 1, file))
    {
        verbose("Failed to open file : \"%s\", failed to read header\n", path);
        fclose(file);
        return NULL;
    }
    Image *image = processbmpheader(buffer, &bpp);
    if(image == NULL)
    {
        verbose("Failed to open file : \"%s\", invalid header\n", path);
        fclose(file);
        return NULL;
    }
    verbose("Image size : %i * %i\n", image->w, image->h);
    fseek(file, offset, SEEK_SET);
    switch (bpp)
    {
    case 24:
        offset = loadbmp_24(file, image);
        break;
    case 32:
        offset = loadbmp_32(file, image, false);
        break;
    case 33:
        offset = loadbmp_32(file, image, true);
        break;
    case 16:
        offset = loadbmp_16(file, image);
        break;
    case 8:
        offset = loadbmp_8(file, image);
        break;
    case 1:
        offset = loadbmp_1(file, image);
        break;
    default:
        printf("Warning : unsupported bpp : %u", bpp);
        offset = false;
    }
    fclose(file);
    if(!offset)
    {
        verbose("Failed to load : \"%s\", unexpected file end\n", path);
        free_image(image);
        return NULL;
    }

    return image;
}

OCRAPI static bool loadbmp_32(FILE *file, Image *image, bool rgba)
{
    USHORT x;
    USHORT y;
    Pixel p;
    for(y = image->h - 1;; y--)
    {
        for(x = 0; x < image->w; x++)
        {
            if(!rgba)
                fseek(file, 1, SEEK_CUR);
            if(!fread(&p, sizeof(Pixel), 1, file))
            {
                return false;
            }
            setpixel(image, x, y, p);
            if(rgba)
                fseek(file, 1, SEEK_CUR);
        }
        if(y == 0)
            break;
    }
    return true;
}

OCRAPI static bool loadbmp_24(FILE *file, Image *image)
{
    USHORT x;
    USHORT y;
    USHORT w = (3 * (image->w + 1) / 4) * 4;
    Pixel p;
    for(y = image->h - 1;; y--)
    {
        for(x = 0; x < image->w; x++)
        {
            if(!fread(&p, sizeof(Pixel), 1, file))
            {
                return false;
            }
            setpixelrgb(image, x, y, p.blue, p.green, p.red);
        }
        fseek(file, w - (x * 3), SEEK_CUR);
        if(y == 0)
            break;
    }
    return true;
}

OCRAPI static bool loadbmp_16(FILE *file, Image *image)
{
    USHORT x;
    USHORT y;
    USHORT w = (2 * (image->w + 1) / 4) * 4;
    USHORT b;
    USHORT rmask = 0xF800;
    for(y = image->h - 1; ; y++)
    {
        for(x = 0; x < image->w; x++)
        {
            if(!fread(&b, 2, 1, file))
            {
                return false;
            }
            setpixelrgb(image, x, y, (b & 0x1F), (b & 0x7E0) >> 5, (b & rmask) >> 11);
        }
        fseek(file, w - (x * 2), SEEK_CUR);
        if(y == 0)
            break;
    }
    return image;
}

OCRAPI static bool loadbmp_8(FILE *file, Image *image)
{
    printf("Warning 8 byte per pixel bitmaps are only supported with RRRGGGBB color set");
    USHORT x;
    USHORT y;
    USHORT w = (3 * (image->w + 1) / 4) * 4;
    BYTE b;
    for(y = image->h - 1;; y--)
    {
        for(x = 0; x < image->w; x++)
        {
            if(!fread(&b, 1, 1, file))
            {
                return false;
            }
            setpixelrgb(image, x, y, (b & 0x7) * 32, ((b & 0x38) >> 3) * 32, (b >> 6) * 64);
        }
        fseek(file, w - (x * 3), SEEK_CUR);
        if(y == 0)
            break;
    }
    return true;
}

OCRAPI static bool loadbmp_1(FILE *file, Image *image)
{
    USHORT x;
    USHORT y;
    BYTE i;
    BYTE b;
    for(y = image->h - 1;; y--)
    {
        for(x = 0; x < image->w; x++)
        {
            if(!fread(&b, 1, 1, file))
            {
                return false;
            }
            for(i = 0; i < 8 && x < image->w; i++)
            {
                if(b & 1)
                    setpixel(image, x, y, (Pixel)
                {
                    255, 255, 255
                });
                else
                    setpixel(image, x, y, (Pixel)
                {
                    0, 0, 0
                });
                b >>= 1;
            }
        }
        if(y == 0)
            break;
    }
    return true;
}

bool save_image(Image *img, const char *path)
{
    if(!canwrite(path))
        return false;
    FILE *file = fopen (path, "wb");
    if (file == NULL)
    {
        verbose("Failed to write file : \"%s\"\n", path);
        return false;
    }
    USHORT x;
    USHORT y;

    BmpHeader bmph;
    USHORT linesize = (3 * (img->w + 1) / 4) * 4;
    BYTE line[linesize];
    memcpy(bmph.bfType, "BM", 2);
    bmph.bfOffBits = 54;
    bmph.bfSize = bmph.bfOffBits + linesize * img->h;
    bmph.bfReserved = 0;
    bmph.biSize = 40;
    bmph.biWidth = img->w;
    bmph.biHeight = img->h;
    bmph.biPlanes = 1;
    bmph.biBitCount = 24;
    bmph.biCompression = 0;
    bmph.biSizeImage =  linesize * img->h;
    bmph.biXPelsPerMeter = 0;
    bmph.biYPelsPerMeter = 0;
    bmph.biClrUsed = 0;
    bmph.biClrImportant = 0;

    if(!(fwrite(&bmph.bfType, 2, 1, file) &&
            fwrite(&bmph.bfSize, 4, 1, file) &&
            fwrite(&bmph.bfReserved, 4, 1, file) &&
            fwrite(&bmph.bfOffBits, 4, 1, file) &&
            fwrite(&bmph.biSize, 4, 1, file) &&
            fwrite(&bmph.biWidth, 4, 1, file) &&
            fwrite(&bmph.biHeight, 4, 1, file) &&
            fwrite(&bmph.biPlanes, 2, 1, file) &&
            fwrite(&bmph.biBitCount, 2, 1, file) &&
            fwrite(&bmph.biCompression, 4, 1, file) &&
            fwrite(&bmph.biSizeImage, 4, 1, file) &&
            fwrite(&bmph.biXPelsPerMeter, 4, 1, file) &&
            fwrite(&bmph.biYPelsPerMeter, 4, 1, file) &&
            fwrite(&bmph.biClrUsed, 4, 1, file) &&
            fwrite(&bmph.biClrImportant, 4, 1, file)))
    {
        verbose("fwrite failed while writting : \"%s\",  %i\n", path, ferror(file));
        fclose(file);
        return false;
    }
    ZeroMemory(line, linesize);
    for (y = img->h - 1;; y--)
    {
        for (x = 0; x < img->w; x++)
        {
            line[3 * x] = getpixel(img, x, y).blue;
            line[3 * x + 1] = getpixel(img, x, y).green;
            line[3 * x + 2] = getpixel(img, x, y).red;
        }
        if(fwrite(line, linesize, 1, file) != 1)
        {
            verbose("fwrite failed while writting : \"%s\", %i\n", path, ferror(file));
            fclose(file);
            return false;
        }
        if(y == 0)
            break;
    }
    fclose(file);
    return true;
}

OCRAPI void setoverwrite(bool b)
{
    boverwrite = b;
}
