#include "ocrdefs.h"
#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>
#include <string.h>
#include <math.h>

#include "textextraction.h"
#include "neurone.h"
#include "filters.h"

static void usage();
static void info();
int main(int argc, char *argv[])
{
    int opt = 0;
    int option_index = 0;
    BYTE operation = OPERATION_OCR;
    bool colors = false;
    bool force = false;
    float angle = 0;
    int bin_type = THRESHOLD;
    int threshold = 127;
    char *output = NULL;

    struct option long_options[] =
    {
        {"help",	no_argument,		0,		'h'},
        {"binarization", required_argument,	0,		0},
        {"binarize", no_argument,		0,	'b'},
        {"force",	no_argument,		0,	'f'},
        {"output",	required_argument,	0,		'o'},
        {"verbose",	no_argument,		0,		'v'},
        {"rotate", required_argument, 0, 'r'},
        {"colors", no_argument, 0, 'c'},
        {"xor forced", required_argument, 0, 'x'},
        {"xor_default", no_argument, 0, 'd'},
        {"rlsa", required_argument, 0, 's'},
        {"rlsa", required_argument, 0, 'z'},
        {"infos", required_argument, 0, 'i'},
        {0, 0, 0, 0}
    };

    if (argc <= 1)
    {
        usage();
        return 0;
    }

    while ((opt = getopt_long(argc, argv, "o:r:s:z:x:dfvcbhi",
                              long_options, &option_index)) != -1)
    {
        switch (opt)
        {
        case 0:
        {
            if (!strcmp(long_options[option_index].name, "binarization"))
            {
                if (!strcmp(optarg, "otsu") || !strcmp(optarg, "otsu1"))
                    bin_type = OTSU1;
                else if (!strcmp(optarg, "otsu2"))
                    bin_type = OTSU2;
                else
                {
                    threshold = atoi(optarg);
                    bin_type = THRESHOLD;
                }
            }
            break;
        }

        case 'h':
            usage();
            return 0;

        case 'o':
            if(!optarg)
                break;
            output = malloc(sizeof(char) * strlen(optarg) + 1);
            sprintf(output, "%s", optarg);
            break;

        case 'v':
            setverbose(1);
            break;

        case 'x':
            if (!optarg) return 0;
            else xor(atoi(optarg));
            return 0;
        case 'd':
            xor_default();
            return 0;
        case 'r':
            if(!optarg)
                break;
            operation = OPERATION_ROTATE;
            angle = atof(optarg);
            break;
        case 'b':
            operation = OPERATION_BINARIZE;
            break;
        case 'f':
            force = true;
            break;
        case 'c':
            colors = true;
            break;
        case 's':
            if(!optarg)
                break;
            setrlsa_radius(atoi(optarg));
            break;
        case 'z':
            if(!optarg)
                break;
            forcerlsa_radius(atoi(optarg));
            break;
        case 'i':
            info();
            return 0;
        default:
            break;
        }
    }
    setoverwrite(force);
    if (argv[optind] == NULL)
    {
        printf("No input image given, exiting\n");
        return 3;
    }

    const char *filename = argv[optind];

    if (output == NULL)
    {
        output = malloc(sizeof(char) * strlen("out.bmp") + 1);
        sprintf(output, "out.bmp");
        printf("No output given, \"%s\" will be used\n", output);
    }

    Image *img = load_image(filename);
    if(img == NULL)
    {
        printf("Failed to load : \"%s\"\n", filename);
        return 2;
    }

    else if (operation == OPERATION_BINARIZE)
    {
        Bin_Image *bimg = binarize(img, bin_type, threshold);
        free_image(img);
        img = bimage_tosurface(bimg);
        free_bimage(bimg);
    }
    else if(operation == OPERATION_ROTATE)
    {
        rotate_image(img, (angle * M_PI) / 180, img->w / 2, img->h / 2);
    }
    else
    {
        detect_anddraw(img, colors, bin_type, threshold);
    }

    if(!save_image(img, output))
    {
        printf("Failed to save : \"%s\"\n", output);
    }

    free_image(img);
    free(output);
    return 0;
}

static void usage()
{
    const char *usagemsg =
        "Usage: mediOCR [options] file...\n"
        "Options:\n"
        "  -h, --help\t\tDisplay this information\n"
        "  --binarization <TYPE>\tuse the TYPE method to binarize the image;\n"
        "                     \tTYPE can be 'otsu{1, 2}' or a number for a "
        "fixed threshold\n"
        "  -b, --binarize-only\tuse this flag to only apply a binarization and "
        "exit\n"
        "  -f, --force\t\tDo not prompt before overwriting\n"
        "  -o, --output <file>\tPlace output into <file>\n"
        "  -v, --verbose\t\tPrint more informations on stdout\n"
        "  -x\t\t\tDisplay 3500 XOR learning\n"
        "  -d  --n\t\tDisplay n XOR learning\n"
        "  -r, --rotate <angle> \tApplies rotation of specified angle in degrees\n"
        "  -c, --colors \t\tActivates color management\n"
        "  -s, --rlsa <radius>\tSets primary rlsa radius\n"
        "  -w, --rsla-forced\tForced rlsa secondary radius\n";
    puts(usagemsg);
}

static void info()
{
    const char *infos =
        "MediOCR 1.0\n"
        "\tBuild time : "__DATE__", "__TIME__"\n"
        "Authors : \n"
        "\tBardoux Antoine\n"
        "\tBarroyer Thibaut\n"
        "\tBoulay Pierre\n"
        "\tLamatte Valentin\n"
        "For more infos, please visit : http://mediocr.net\n";
    puts(infos);
}
